import argparse
import csv
import json
from .core import IXBRL


def main():

    parser = argparse.ArgumentParser(description='Extract financial data from a IXBRL file')
    parser.add_argument('infile',
                        help='file to open and convert',
                        type=argparse.FileType('r')
                        )
    parser.add_argument('--outfile',
                        default="-",
                        help='Where to output the file',
                        type=argparse.FileType('w', encoding='UTF-8')
                        )
    parser.add_argument('--format',
                        choices=['csv', 'json', 'jsonlines', 'jsonl'],
                        default="csv",
                        help='format of the output'
                        )
    parser.add_argument('--fields',
                        choices=['numeric', 'nonnumeric', 'all'],
                        default="all",
                        help='Which fields to output'
                        )

    args = parser.parse_args()

    x = IXBRL(args.infile)

    if args.format == 'csv':
        values = x.to_table(args.fields)
        columns = {}
        for r in values:
            columns = {**dict.fromkeys(r.keys())}
        writer = csv.DictWriter(args.outfile, columns.keys())
        writer.writeheader()
        writer.writerows(values)
    elif args.format in ['jsonlines', 'jsonl']:
        values = x.to_table(args.fields)
        for v in values:
            json.dump(v, args.outfile)
            args.outfile.write("\n")
    elif args.format == "json":
        json.dump(x.to_json(), args.outfile, indent=4)


if __name__ == "__main__":
    main()



from bs4 import BeautifulSoup

from ixbrlparse.components import ixbrlContext, ixbrlNonNumeric, ixbrlNumeric


class IXBRL():

    def __init__(self, f, raise_on_error=True):
        self.soup = BeautifulSoup(f.read(), "xml")
        self.raise_on_error = raise_on_error
        self.errors = []
        self._get_schema()
        self._get_contexts()
        self._get_units()
        self._get_nonnumeric()
        self._get_numeric()

    @classmethod
    def open(cls, filename, raise_on_error=True):
        with open(filename) as a:
            return cls(a, raise_on_error=raise_on_error)

    def _get_schema(self):
        self.schema = self.soup.find(
            ['link:schemaRef', 'schemaRef']).get('xlink:href')
        self.namespaces = {}
        for k in self.soup.find('html').attrs:
            if k.startswith("xmlns") or ":" in k:
                self.namespaces[k] = self.soup.find('html')[k].split(" ")

    def _get_contexts(self):
        self.contexts = {}
        resources = self.soup.find(['ix:resources', 'resources'])
        for s in resources.find_all(['xbrli:context', 'context']):
            self.contexts[s['id']] = ixbrlContext(**{
                "_id": s['id'],
                "entity": {
                    "scheme": s.find(['xbrli:identifier', 'identifier'])['scheme'] if s.find(['xbrli:identifier', 'identifier']) else None,
                    "identifier": s.find(['xbrli:identifier', 'identifier']).text if s.find(['xbrli:identifier', 'identifier']) else None,
                },
                "segments": [
                    {
                        "tag": x.name,
                        "value": x.text.strip(),
                        **x.attrs
                    }
                    for x in s.find(['xbrli:segment', 'segment']).findChildren()
                ] if s.find(['xbrli:segment', 'segment']) else None,
                "instant": s.find(['xbrli:instant', 'instant']).text if s.find(['xbrli:instant', 'instant']) else None,
                "startdate": s.find(['xbrli:startDate', 'startDate']).text if s.find(['xbrli:startDate', 'startDate']) else None,
                "enddate": s.find(['xbrli:endDate', 'endDate']).text if s.find(['xbrli:endDate', 'endDate']) else None,
            })

    def _get_units(self):
        self.units = {}
        resources = self.soup.find(['ix:resources', 'resources'])
        for s in resources.find_all(['xbrli:unit', 'unit']):
            self.units[s['id']] = s.find(
                ['xbrli:measure', 'measure']
            ).text if s.find(
                ['xbrli:measure', 'measure']
            ) else None

    def _get_nonnumeric(self):
        self.nonnumeric = []
        for s in self.soup.find_all({'nonNumeric'}):
            element = {
                "context": self.contexts.get(s['contextRef'], s['contextRef']),
                "name": s['name'],
                "format_": s.get('format'),
                "value": s.text.strip().replace("\n", "")
            }
            try:
                self.nonnumeric.append(ixbrlNonNumeric(**element))
            except Exception as e:
                self.errors.append({
                    'error': e,
                    'element': s,
                })
                if self.raise_on_error:
                    raise

    def _get_numeric(self):
        self.numeric = []
        for s in self.soup.find_all({'nonFraction'}):
            element = {
                "text": s.text,
                "context": self.contexts.get(
                    s['contextRef'],
                    s['contextRef']
                ),
                "unit": self.units.get(s['unitRef'], s['unitRef']),
                **s.attrs
            }
            try:
                self.numeric.append(
                    ixbrlNumeric(element)
                )
            except Exception as e:
                self.errors.append({
                    'error': e,
                    'element': s,
                })
                if self.raise_on_error:
                    raise

    def to_json(self):
        return {
            "schema": self.schema,
            "namespaces": self.namespaces,
            "contexts": {c: ct.to_json() for c, ct in self.contexts.items()},
            "units": self.units,
            "nonnumeric": [a.to_json() for a in self.nonnumeric],
            "numeric": [a.to_json() for a in self.numeric],
            "errors": len(self.errors),
        }

    def to_table(self, fields="numeric"):
        if fields == 'nonnumeric':
            values = self.nonnumeric
        elif fields == "numeric":
            values = self.numeric
        else:
            values = self.nonnumeric + self.numeric

        ret = []
        for v in values:
            if v.context.segments:
                segments = {
                    "segment:{}".format(i): "{} {} {}".format(
                        s.get("tag", ""), s.get("dimension"), s.get("value")
                    ).strip() for i, s in enumerate(v.context.segments)
                }
            else:
                segments = {"segment:0": ""}

            ret.append({
                "schema": " ".join(
                    self.namespaces.get(
                        "xmlns:{}".format(v.schema),
                        [v.schema]
                    )
                ),
                "name": v.name,
                "value": v.value,
                "unit": v.unit if hasattr(v, 'unit') else None,
                "instant": str(v.context.instant)
                if v.context.instant else None,
                "startdate": str(v.context.startdate)
                if v.context.startdate else None,
                "enddate": str(v.context.enddate)
                if v.context.enddate else None,
                **segments
            })
        return ret



from .transform import get_format


class ixbrlNumeric:

    # contextref
    # decimals
    # format
    # name
    # scale
    # sign
    # text
    # unitref
    # xmlns:ix
    def __init__(self, attrs):
        name = attrs.get('name', "").split(":", maxsplit=1)
        if len(name) == 2:
            self.schema = name[0]
            self.name = name[1]
        else:
            self.schema = 'unknown'
            self.name = name[0]

        self.text = attrs.get('value', attrs.get('text'))
        self.context = attrs.get('context')
        self.unit = attrs.get('unit')

        format_ = {
            "format_": attrs.get('format'),
            "decimals": attrs.get('decimals', "0"),
            "scale": attrs.get('scale', 0),
            "sign": attrs.get('sign', ""),
        }
        self.format = get_format(format_['format_'])(**format_)

        try:
            self.value = self.format.parse_value(self.text)
        except ValueError:
            print(attrs)
            raise

    def to_json(self):
        values = self.__dict__
        values['context'] = self.context.to_json()
        return values



class ixbrlNonNumeric:

    def __init__(self, context, name, format_, value):

        name = name.split(":", maxsplit=1)
        if len(name) == 2:
            self.schema = name[0]
            self.name = name[1]
        else:
            self.schema = 'unknown'
            self.name = name[0]

        self.context = context
        self.format = format_
        self.value = value

    def to_json(self):
        values = self.__dict__
        values['context'] = self.context.to_json()
        return values



import datetime


class ixbrlContext:

    def __init__(self, _id, entity, segments, instant, startdate, enddate):
        self.id = _id
        self.entity = entity
        self.segments = segments
        # @TODO: parse dates here
        self.instant = datetime.datetime.strptime(
            instant.strip(), "%Y-%m-%d").date() if instant else None
        self.startdate = datetime.datetime.strptime(
            startdate.strip(), "%Y-%m-%d").date() if startdate else None
        self.enddate = datetime.datetime.strptime(
            enddate.strip(), "%Y-%m-%d").date() if enddate else None

    def __repr__(self):
        if self.startdate and self.enddate:
            datestr = "{} to {}".format(self.startdate, self.enddate)
        else:
            datestr = str(self.instant)

        segmentstr = " (with segments)" if self.segments else ""

        return "<IXBRLContext {} [{}]{}>".format(self.id, datestr, segmentstr)

    def to_json(self):
        values = self.__dict__
        for i in ['startdate', 'enddate', 'instant']:
            if isinstance(values[i], datetime.date):
                values[i] = str(values[i])
        return values



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="ixbrlparse",
    packages=find_packages(),
    version="0.1.2",
    author="David Kane",
    author_email="david@dkane.net",
    description="A python module for getting useful data out of ixbrl files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/drkane/ixbrl-parse",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    install_requires=[
        'beautifulsoup4',
        'lxml',
    ],
)



class ixbrlFormat:

    def __init__(self, format_, decimals, scale, sign):

        if decimals.lower() == "inf":
            self.decimals = None
        else:
            self.decimals = int(decimals)

        self.format = None
        if format_:
            format_ = format_.split(":")
            if len(format_) > 1:
                self.format = ":".join(format_[1:])
                self.namespace = format_[0]
            else:
                self.format = ":".join(format_)
                self.namespace = None

        self.scale = int(scale)
        self.sign = sign

    def parse_value(self, value):

        if isinstance(value, (int, float)):
            return value

        if value in ('-', ''):
            return 0

        value = value.replace(' ', '')
        value = value.replace(',', '')
        value = float(value)

        if self.sign == "-":
            value = value * -1

        if self.scale != 0:
            value = value * (10 ** self.scale)

        return value


class ixtZeroDash(ixbrlFormat):

    def parse_value(self, value):
        return 0


class ixtNoContent(ixbrlFormat):

    def parse_value(self, value):
        return 0


class ixtNumComma(ixbrlFormat):

    def parse_value(self, value):
        value = value.replace('.', '')
        value = value.replace(',', '.')
        return super().parse_value(value)


def get_format(format_):

    if format_ is None:
        return ixbrlFormat

    format_ = format_.split(":")
    if len(format_) > 1:
        namespace = format_[0]
        format_ = ":".join(format_[1:])
    else:
        namespace = None
        format_ = ":".join(format_)

    format_ = format_.replace("-", "")

    if format_ in ('zerodash', 'numdash'):
        return ixtZeroDash

    if format_ == 'nocontent':
        return ixtNoContent

    if format_ in ('numdotdecimal', 'numcommadot', 'numspacedot'):
        return ixbrlFormat

    if format_ in ('numcomma', 'numdotcomma', 'numspacecomma', 'numcommadecimal'):
        return ixtNumComma

    raise NotImplementedError(
        'Format "{}" not implemented (namespace "{}")'.format(
            format_,
            namespace,
        )
    )
